--- qmail-send.c	Mon Jun 15 12:53:16 1998
+++ qmail-send2.c	Thu Mar 7 23:28:08 2002
@@ -36,6 +36,7 @@
 /* critical timing feature #2: if triggered, respond within fixed time */
 /* important timing feature: when triggered, respond instantly */
 #define SLEEP_TODO 1500 /* check todo/ every 25 minutes in any case */
+#define ONCEEVERY 10 /* run todo maximal once every N seconds */
 #define SLEEP_FUZZ 1 /* slop a bit on sleeps to avoid zeno effect */
 #define SLEEP_FOREVER 86400 /* absolute maximum time spent in select() */
 #define SLEEP_CLEANUP 76431 /* time between cleanups */
@@ -1215,7 +1216,7 @@
 
 /* this file is too long ---------------------------------------------- TODO */
 
-datetime_sec nexttodorun;
+datetime_sec nexttodorun, lasttodorun;
 DIR *tododir; /* if 0, have to opendir again */
 stralloc todoline = {0};
 char todobuf[SUBSTDIO_INSIZE];
@@ -1225,6 +1226,7 @@
 void todo_init()
 {
  tododir = 0;
+ lasttodorun = now();
  nexttodorun = now();
  trigger_set();
 }
@@ -1265,6 +1267,39 @@
 
  if (flagexitasap) return;
 
+ /*
+  * qmail normally does a queue run for every injected message and then
+  * schedules it for local/remote delivery.
+  * If we get too many injections the queue runs will stall scheduling
+  * of local/remote deliveries. Usually we see a sharp drop in concurrent
+  * deliveries while new messages arrive at high rate in the queue.
+  * In such a situation it doesn't make sense to process the queue after
+  * each and every injection. It is more important to fill the delivery
+  * slots with work to drain the queue.
+  * This patch will allow for only one queue run every N seconds. It
+  * remembers the last complete queue run and will just return if we
+  * get triggered before last queue run plus N seconds. If we get
+  * triggered before N seconds it will schedule a forced queue run at
+  * last complete queue run plus N seconds to not miss the new additions
+  * to the queue.
+  * Some more thoughts:
+  *  - This is proof of concept and should be made at least dynamically
+  *    configureable through a control file (and reread with -HUP)
+  *  - This could even be made selftuning by measuring the time it takes
+  *    to do a queue run and do the run delay only after a certain
+  *    threshold has been reached
+  *  - the optimal values for delaying have to found
+  *
+  *  20011108, Initial revision, opi@nrg4u.com
+  */
+
+ /* run todo maximal once every N seconds */
+ if (recent < (lasttodorun + ONCEEVERY))
+  {
+   nexttodorun = lasttodorun + ONCEEVERY; /* do this to wake us up in N secs */
+   return; /* skip todo run this time */
+  }
+
  if (!tododir)
   {
    if (!trigger_pulled(rfds))
@@ -1277,6 +1312,7 @@
      pausedir("todo");
      return;
     }
+   lasttodorun = recent;
    nexttodorun = recent + SLEEP_TODO;
   }
 
